@@ -1,4 +1,4 @@
-## routines for very large dataset generalized additive modelling.
+# routines for very large dataset generalized additive modelling.
 ## (c) Simon N. Wood 2009-2015
 
 
@@ -23,7 +23,7 @@ rwMatrix <- function(stop,row,weight,X,trans=FALSE) {
   if (is.matrix(X)) { n <- nrow(X);p<-ncol(X);ok <- TRUE} else { n<- length(X);p<-1;ok<-FALSE}
   stop <- stop - 1;row <- row - 1 ## R indices -> C indices
   oo <-.C(C_rwMatrix,as.integer(stop),as.integer(row),as.double(weight),X=as.double(X),
-          as.integer(n),as.integer(p),trans=as.integer(trans))
+          as.integer(n),as.integer(p),trans=as.integer(trans),work=as.double(rep(0,n*p)))
   if (ok) return(matrix(oo$X,n,p)) else
   return(oo$X) 
 } ## rwMatrix
@@ -122,7 +122,8 @@ compress.df <- function(dat,m=NULL) {
   } else {
     mm <- mm * m 
   } 
-  xu <- uniquecombs(as.matrix(dat))
+  #xu <- uniquecombs(as.matrix(dat))
+  xu <- uniquecombs(dat)
   if (nrow(xu)>mm*mf) { ## too many unique rows to use only unique
     for (i in 1:d) if (!is.factor(dat[,i])) { ## round the metric variables
       xl <- range(dat[,i])
@@ -131,7 +132,8 @@ compress.df <- function(dat,m=NULL) {
       kx <- round((dat[,i]-xl[1])/dx)+1
       dat[,i] <- xu[kx] ## rounding the metric variables
     }
-    xu <- uniquecombs(as.matrix(dat))
+    #xu <- uniquecombs(as.matrix(dat))
+    xu <- uniquecombs(dat)
   }  
   k <- attr(xu,"index")
   ## shuffle rows in order to avoid induced dependencies between discretized
@@ -153,76 +155,163 @@ compress.df <- function(dat,m=NULL) {
   xu[ii,] <- xu  ## shuffle rows of xu
   k <- ii[k]     ## correct k index accordingly
   ## ... finished shuffle
-  xu <- as.data.frame(xu)
+  #xu <- as.data.frame(xu)
   ## set names and levels to match dat...
-  names(xu) <- names(dat)
-  for (i in 1:d) if (is.factor(dat[,i])) {
-    xu[,i] <- as.factor(xu[,i])
-    levels(xu[,i]) <- levels(dat[,i])
-  }
+  #names(xu) <- names(dat)
+  #for (i in 1:d) if (is.factor(dat[,i])) {
+  #  xu[,i] <- as.factor(xu[,i])
+  #  levels(xu[,i]) <- levels(dat[,i])
+  #}
   k -> attr(xu,"index")
   xu
 } ## compress.df
 
-discrete.mf <- function(gp,mf,pmf,m=NULL) {
+check.term <- function(term,rec) {
+## utility function for discrete.mf. Checks whether variables in "term"
+## have already been discretized, and if so whether this discretization 
+## can be re-used for the current "term". Stops if term already discretized
+## but we can't re-use discretization. Otherwise returns index of k index
+## or 0 if the term is not in the existing list.
+  ii <- which(rec$vnames%in%term)
+  if (length(ii)) { ## at least one variable already discretized
+    if (length(term)==rec$d[min(ii)]) { ## dimensions match previous discretization
+      if (sum(!(term%in%rec$vnames[ii]))) ("bam can not discretize with this nesting structure")
+      else return(rec$ki[min(ii)]) ## all names match previous - retun index of previous
+    } else stop("bam can not discretize with this nesting structure")
+  } else return(0) ## no match
+} ## check.term
+
+discrete.mf <- function(gp,mf,pmf,m=NULL,full=TRUE) {
 ## discretize the covariates for the terms specified in smooth.spec
 ## id and factor by not allowed. pmf is a model frame for just the 
 ## parametric terms --- mini.mf is applied to this.
+## if full is FALSE then parametric and response terms are ignored
+## and what is returned is a list where columns can be of 
+## different lengths.
 ## NOTE: matrix arguments not allowed but not caught yet.
   mf0 <- list()
   nk <- 0 ## count number of index vectors to avoid too much use of cbind
-  for (i in 1:length(gp$smooth.spec)) nk <- nk + 
+  for (i in 1:length(gp$smooth.spec)) nk <- nk + as.numeric(gp$smooth.spec[[i]]$by!="NA") +
     if (inherits(gp$smooth.spec[[i]],"tensor.smooth.spec")) length(gp$smooth.spec[[i]]$margin) else 1
-  k <- matrix(0,nrow(mf),nk)
+  k <- matrix(0,nrow(mf),nk) ## each column is an index vector
   ik <- 0 ## index counter
   nr <- rep(0,nk) ## number of rows for term
+  ## structure to record terms already processed...
+  rec <- list(vnames = rep("",0), ## variable names
+              ki = rep(0,0),      ## index of original index vector var relates to  
+              d = rep(0,0))       ## dimension of terms involving this var
   ## loop through the terms discretizing the covariates...
   for (i in 1:length(gp$smooth.spec)) {
     mi <- if (is.null(m)||length(m)==1) m else m[i]
+    if (!is.null(gp$smooth.spec[[i]]$id)) stop("discretization can not handle smooth ids")
+    ## deal with any by variable (should always go first as basically a 1D marginal)... 
+    if (gp$smooth.spec[[i]]$by!="NA") {
+      ##stop("currently discrete methods do not handle by variables")
+      termi <- gp$smooth.spec[[i]]$by ## var name
+      ##if (is.factor(mf[termi])) stop("discretization can not handle factor by variables")
+      ik.prev <- check.term(termi,rec) ## term already discretized?
+      ik <- ik + 1 ## increment index counter
+      if (ik.prev==0) { ## new discretization required
+         mfd <- compress.df(mf[termi],m=mi)
+         k[,ik] <- attr(mfd,"index")
+         nr[ik] <- nrow(mfd)
+         mf0 <- c(mf0,mfd) 
+         ## record variable discretization info...
+         rec$vnames <- c(rec$vnames,termi)
+         rec$ki <- c(rec$ki,ik)
+         rec$d <- c(rec$d,1)
+       } else { ## re-use an earlier discretization...
+         k[,ik] <- k[,ik.prev]
+         nr[ik] <- nr[ik.prev]
+       }
+    }  
     if (inherits(gp$smooth.spec[[i]],"tensor.smooth.spec")) { ## tensor branch
       for (j in 1:length(gp$smooth.spec[[i]]$margin)) { ## loop through margins
-        mfd <- compress.df(mf[gp$smooth.spec[[i]]$margin[[j]]$term],m=mi)
+        termj <- gp$smooth.spec[[i]]$margin[[j]]$term
+        ik.prev <- check.term(termj,rec) 
         ik <- ik + 1
-        k[,ik] <- attr(mfd,"index")
-        nr[ik] <- nrow(mfd)
-        mf0 <- c(mf0,mfd)
+        if (ik.prev==0) { ## new discretization required
+          mfd <- compress.df(mf[termj],m=mi)
+          k[,ik] <- attr(mfd,"index")
+          nr[ik] <- nrow(mfd)
+          mf0 <- c(mf0,mfd)
+          ## record details...
+          d <- length(termj)
+          rec$vnames <- c(rec$vnames,termj)
+          rec$ki <- c(rec$ki,rep(ik,d))
+          rec$d <- c(rec$d,rep(d,d))
+        } else { ## re-use an earlier discretization... 
+          k[,ik] <- k[,ik.prev]
+          nr[ik] <- nr[ik.prev]
+        }
       }
     } else { ## not te or ti...
-      mfd <- compress.df(mf[gp$smooth.spec[[i]]$term],m=mi)
-      ik <- ik + 1
-      k[,ik] <- attr(mfd,"index")
-      nr[ik] <- nrow(mfd)
-      mf0 <- c(mf0,mfd)
+      termi <- gp$smooth.spec[[i]]$term 
+      ik.prev <- check.term(termi,rec) 
+      ik <- ik + 1 ## index counter
+      if (ik.prev==0) { ## new discretization required
+        mfd <- compress.df(mf[termi],m=mi)
+        k[,ik] <- attr(mfd,"index")
+        nr[ik] <- nrow(mfd)
+        mf0 <- c(mf0,mfd) 
+        d <- length(termi)
+        rec$vnames <- c(rec$vnames,termi)
+        rec$ki <- c(rec$ki,rep(ik,d))
+        rec$d <- c(rec$d,rep(d,d))
+      } else { ## re-use an earlier discretization...
+        k[,ik] <- k[,ik.prev]
+        nr[ik] <- nr[ik.prev]
+      }
     }
-    ## deal with any by variable... 
-    if (gp$smooth.spec[[i]]$by!="NA") {
-      stop("currently discrete methods do not handle by variables")
-      if (is.factor(mf[gp$smooth.spec[[i]]$by])) stop("discretization can not handle factor by variables")
-      if (!is.null(gp$smooth.spec[[i]]$id)) stop("discretization can not handle smooth ids")
-      mf0[[gp$smooth.spec[[i]]$by]] <- rep(1,nr[ik]) ## actual by variables are handled by discrete methods
-    }  
+   
   } ## main term loop
+  ## obtain parametric terms and..
   ## pad mf0 so that all rows are the same length
   ## padding is necessary if gam.setup is to be used for setup
-  maxr <- max(nr) 
-  pmf0 <- mini.mf(pmf,maxr)
-  if (nrow(pmf0)>maxr) maxr <- nrow(pmf0)
-  mf0 <- c(mf0,pmf0)
-  for (i in 1:length(mf0)) {
-    me <- length(mf0[[i]]) 
-    if (me < maxr) mf0[[i]][(me+1):maxr] <- sample(mf0[[i]],maxr-me,replace=TRUE)
-  }
-  ## add response so that gam.setup can do its thing...
-  mf0[[gp$response]] <- sample(mf[[gp$response]],maxr)
+  if (full) {
+    maxr <- max(nr) 
+    pmf0 <- mini.mf(pmf,maxr) ## deal with parametric components
+    if (nrow(pmf0)>maxr) maxr <- nrow(pmf0)
+    mf0 <- c(mf0,pmf0) ## add parametric terms to end of mf0
+    for (i in 1:length(mf0)) {
+      me <- length(mf0[[i]]) 
+      if (me < maxr) mf0[[i]][(me+1):maxr] <- sample(mf0[[i]],maxr-me,replace=TRUE)
+    }
+    ## add response so that gam.setup can do its thing...
+    mf0[[gp$response]] <- sample(mf[[gp$response]],maxr)
 
-  ## mf0 is the discretized model frame (actually a list), padded to have equal length rows
-  ## k is the index vector for each sub-matrix, only the first nr rows of which are
-  ## to be retained... Use of check.names=FALSE ensures, e.g. 'offset(x)' not changed...
-
-  ## now copy back into mf so terms unchanged
-  mf <- mf[1:maxr,]
-  for (na in names(mf0)) mf[[na]] <- mf0[[na]]  
+    ## mf0 is the discretized model frame (actually a list), padded to have equal length rows
+    ## k is the index vector for each sub-matrix, only the first nr rows of which are
+    ## to be retained... Use of check.names=FALSE ensures, e.g. 'offset(x)' not changed...
 
+    ## now copy back into mf so terms unchanged
+    mf <- mf[1:maxr,]
+    for (na in names(mf0)) mf[[na]] <- mf0[[na]]  
+  } else mf <- mf0
+  ## finally one more pass through, expanding k and nr to deal with replication that
+  ## will occur with factor by variables...
+  ik <- ncol(k)+1 ## starting index col for this term in k
+  for (i in length(gp$smooth.spec):1) { ## work down through terms so insertion painless
+    if (inherits(gp$smooth.spec[[i]],"tensor.smooth.spec")) nd <-  
+         length(gp$smooth.spec[[i]]$margin) else nd <- 1 ## number of indices
+    ik <- ik - nd ## starting index if no by  
+    if (gp$smooth.spec[[i]]$by!="NA") {
+      ik <- ik - 1 ## first index
+      nd <- nd + 1 ## number of indices
+      byvar <- mf[[gp$smooth.spec[[i]]$by]]
+      if (is.factor(byvar)) { ## then need to expand nr and index matrix
+        nex <- length(levels(byvar))  ## number of copies of term indices
+        if (is.ordered(byvar)) nex <- nex - 1 ## first level dropped
+        if (nex>0) { ## insert index copies
+          ii0 <- if (ik>1) 1:(ik-1) else rep(0,0) ## earlier
+          ii1 <- if (ik+nd-1 < length(nr)) (ik+nd):length(nr) else rep(0,0) ## later
+          ii <- ik:(ik+nd-1) ## cols for this term
+          nr <- c(nr[ii0],rep(nr[ii],nex),nr[ii1])
+          k <- cbind(k[,ii0,drop=FALSE],k[,rep(ii,nex),drop=FALSE],k[,ii1,drop=FALSE])
+        }
+      } ## factor by 
+    } ## existing by
+  } ## smooth.spec loop
   list(mf=mf,k=k,nr=nr)
 } ## discrete.mf
 
@@ -269,6 +358,7 @@ mini.mf <-function(mf,chunk.size) {
   } else if (is.factor(mf[[j]])) { ## factor variable...
     ## randomly sample one row from each factor level...
     find <- apply(X=as.matrix(levels(mf[[j]])),MARGIN=1,FUN=fun,fac=mf[[j]],ind=ind)
+    find <- find[is.finite(find)] ## in case drop.unused.levels==FALSE, so that there ar levels without rows
     nf <- length(find)
     mf0[(k+1):(k+nf),] <- mf[find,]
     k <- k + nf
@@ -507,7 +597,7 @@ bgam.fitd <- function (G, mf, gp ,scale , coef=NULL,etastart = NULL,
   object$edf2 <- rowSums(Vc*qrx$R)/scale
   object$Vc <- Vc
   object$outer.info <- list(grad = prop$grad,hess=prop$hess)  
-  
+  object$AR1.rho <- rho
   object$R <- pchol(qrx$R,npt)
   piv <- attr(object$R,"pivot") 
   object$R[,piv] <- object$R   
@@ -577,6 +667,10 @@ bgam.fit <- function (G, mf, chunk.size, gp ,scale ,gamma,method, coef=NULL,etas
 
     if (!is.null(cl)&&inherits(cl,"cluster")) {
       n.threads <- length(cl)
+      while(nobs/n.threads < ncol(G$X)) n.threads <- n.threads - 1
+      if (n.threads < length(cl)) { 
+        warning("Too many cluster nodes to use all efficiently")
+      }
     } else n.threads <- 1
 
     if (n.threads>1) { ## set up thread argument lists
@@ -1066,11 +1160,15 @@ pabapr <- function(arg) {
                         na.action=arg$na.action)
 }
 
-predict.bam <- function(object,newdata,type="link",se.fit=FALSE,terms=NULL,
+predict.bam <- function(object,newdata,type="link",se.fit=FALSE,terms=NULL,exclude=NULL,
                         block.size=50000,newdata.guaranteed=FALSE,na.action=na.pass,
-                        cluster=NULL,...) {
+                        cluster=NULL,discrete=TRUE,n.threads=1,...) {
 ## function for prediction from a bam object, possibly in parallel
   ## remove some un-needed stuff from object
+  if (discrete && !is.null(object$dinfo)) {
+    return(predict.bamd(object,newdata,type,se.fit,terms,exclude,
+                        block.size,newdata.guaranteed,na.action,n.threads,...))
+  }
   object$Sl <- object$qrx <- object$R <- object$F <- object$Ve <-
   object$Vc <- object$G <- object$residuals <- object$fitted.values <-
   object$linear.predictors <- NULL
@@ -1084,11 +1182,11 @@ predict.bam <- function(object,newdata,type="link",se.fit=FALSE,terms=NULL,
   if (n < 100*n.threads) n.threads <- 1 ## not worth the overheads
   if (n.threads==1) { ## single threaded call
     if (missing(newdata)) return(
-      predict.gam(object,newdata=object$model,type=type,se.fit=se.fit,terms=terms,
+      predict.gam(object,newdata=object$model,type=type,se.fit=se.fit,terms=terms,exclude=exclude,
                         block.size=block.size,newdata.guaranteed=newdata.guaranteed,
                         na.action=na.action,...)
     ) else return(
-      predict.gam(object,newdata=newdata,type=type,se.fit=se.fit,terms=terms,
+      predict.gam(object,newdata=newdata,type=type,se.fit=se.fit,terms=terms,exclude=exclude,
                         block.size=block.size,newdata.guaranteed=newdata.guaranteed,
                         na.action=na.action,...))
   } else { ## parallel call...
@@ -1099,7 +1197,7 @@ predict.bam <- function(object,newdata,type="link",se.fit=FALSE,terms=NULL,
     for (i in 1:n.threads) { 
       n0 <- n1+1;n1 <- n1+nt[i]
       ind <- n0:n1 ## this thread's data block from mf
-      arg[[i]] <- list(object=object,type=type,se.fit=se.fit,terms=terms,
+      arg[[i]] <- list(object=object,type=type,se.fit=se.fit,terms=terms,exclude=exclude,
                         block.size=block.size,newdata.guaranteed=newdata.guaranteed,
                         na.action=na.action)
       arg[[i]]$object$model <- object$model[1:2,] ## save space
@@ -1159,8 +1257,11 @@ bam.fit <- function(G,mf,chunk.size,gp,scale,gamma,method,rho=0,
 
    if (n>chunk.size) { ## then use QR accumulation approach
      if (!is.null(cl)&&inherits(cl,"cluster")) { 
-       ## require(parallel)
        n.threads <- length(cl)
+       while(n/n.threads < ncol(G$X)) n.threads <- n.threads - 1
+       if (n.threads < length(cl)) { 
+         warning("Too many cluster nodes to use all efficiently")
+       }
      } else n.threads <- 1
 
      G$coefficients <- rep(0,ncol(G$X))
@@ -1422,6 +1523,162 @@ bam.fit <- function(G,mf,chunk.size,gp,scale,gamma,method,rho=0,
    object
 } # end of bam.fit
 
+predict.bamd <- function(object,newdata,type="link",se.fit=FALSE,terms=NULL,exclude=NULL,
+                        block.size=50000,newdata.guaranteed=FALSE,na.action=na.pass,n.threads=1,...) {
+## function for prediction from a bam object, by discrete methods
+  ## remove some un-needed stuff from object
+  object$Sl <- object$qrx <- object$R <- object$F <- object$Ve <-
+  object$Vc <- object$G <- object$residuals <- object$fitted.values <-
+  object$linear.predictors <- NULL
+  gc()
+  if (missing(newdata)) newdata <- object$model
+
+  if (type=="iterms") {
+    type <- "terms"
+    warning("iterms reset to terms")
+  }
+  
+  if (!is.null(exclude)) warning("exclude ignored by discrete prediction at present")
+
+  newdata <- predict.gam(object,newdata=newdata,type="newdata",se.fit=se.fit,terms=terms,exclude=exclude,
+            block.size=block.size,newdata.guaranteed=newdata.guaranteed,
+            na.action=na.action,...) 
+
+  ## Parametric terms have to be dealt with safely, but without forming all terms 
+  ## or a full model matrix. Strategy here is to use predict.gam, having removed
+  ## key smooth related components from model object, so that it appears to be
+  ## a parametric model... 
+  offset <- 0
+  if (object$nsdf) { ## deal with parametric terms...
+    ## save copies of smooth info...
+    smooth <- object$smooth; coef <- object$coefficients; Vp <- object$Vp
+    ## remove key smooth info from object 
+    object$coefficients <-  object$coefficients[1:object$nsdf]
+    object$Vp <- object$V[1:object$nsdf,1:object$nsdf]
+    object$smooth <- NULL
+    ## get prediction for parametric component. Always "lpmatrix", unless terms required.
+    ptype <- if (type %in% c("terms","iterms")) type else "lpmatrix"
+    pp <- predict.gam(object,newdata=newdata,type=ptype,se.fit=se.fit,terms=terms,exclude=exclude,
+            block.size=block.size,newdata.guaranteed=TRUE,
+            na.action=na.action,...)  
+    ## restore smooths to 'object'
+    object$coefficients <- coef
+    object$Vp <- Vp
+    object$smooth <- smooth
+    if (ptype=="lpmatrix") {
+      offset <- attr(pp,"model.offset")
+      if (is.null(offset)) offset <- 0
+    }
+  } ## parametric component dealt with
+
+  ## now discretize covariates... 
+  dk <- discrete.mf(object$dinfo$gp,mf=newdata,pmf=NULL,full=FALSE)
+
+  Xd <- list() ### list of discrete model matrices...
+  if (object$nsdf>0) {
+     Xd[[1]] <- if (type%in%c("term","iterms")) matrix(0,0,0) else pp 
+     kd <- cbind(1:nrow(newdata),dk$k) ## add index for parametric part to index list
+     kb <- k <- 2; 
+     dk$nr <- c(NA,dk$nr) ## need array index to match elements of Xd
+  } else {
+    kb <- k <- 1; 
+  }
+  
+  ts <- object$dinfo$ts
+  dt <- object$dinfo$dt   
+  for (i in 1:length(object$smooth)) { ## work through the smooth list
+    ## first deal with any by variable (as first marginal of tensor)...
+    if (object$smooth[[i]]$by!="NA") {
+      by.var <- dk$mf[[object$smooth[[i]]$by]][1:dk$nr[k]]
+      if (is.factor(by.var)) { 
+         ## create dummy by variable...
+         by.var <- as.numeric(by.var==object$smooth[[i]]$by.level)  
+      }
+      Xd[[k]] <- matrix(by.var,dk$nr[k],1)
+      k <- k + 1
+    }
+    ## ... by done
+    if (inherits(object$smooth[[i]],"tensor.smooth")) { 
+      nmar <- length(object$smooth[[i]]$margin) 
+      if (!is.null(object$smooth[[i]]$rind)) {
+         ## terms re-ordered for efficiency, so the same has to be done on indices...
+         rind <- k:(k+dt[kb]-1) ## could use object$dinfo$dt[kb]   
+         dk$nr[rind] <- dk$nr[k+object$smooth[[i]]$rind-1]
+         kd[,rind] <- kd[,k+object$smooth[[i]]$rind-1]
+      } 
+      XP <- object$smooth[[i]]$XP         
+      for (j in 1:nmar) {
+        Xd[[k]] <- PredictMat(smooth[[i]]$margin[[j]],dk$mf,n=dk$nr[k])
+        if (!is.null(XP)&&(j<=length(XP))&&!is.null(XP[[j]])) Xd[[k]] <- Xd[[k]]%*%XP[[j]]
+        k <- k + 1 
+      }
+    } else { ## not a tensor smooth
+      object$smooth[[i]]$by <- "NA" ## have to ensure by not applied here!
+      Xd[[k]] <- PredictMat(object$smooth[[i]],dk$mf,n=dk$nr[k])
+      k <- k + 1
+    }
+    kb <- kb + 1
+  }
+   
+  ## end of discrete set up
+  se <- se.fit
+  if (type=="terms") {
+    if (object$nsdf>0) {
+      if (se) {
+        fit <- cbind(pp$fit,matrix(0,nrow(kd),length(object$smooth)))
+        se.fit <- cbind(pp$se.fit,matrix(0,nrow(kd),length(object$smooth))) 
+      } else fit <- cbind(pp,matrix(0,nrow(kd),length(object$smooth)))
+      k <- 2; ## starting Xd
+      kk <- ncol(fit) - length(object$smooth) + 1 ## starting col of fit for smooth terms
+    } else {
+      if (se) {
+        fit <- matrix(0,nrow(kd),length(object$smooth))
+        se.fit <- matrix(0,nrow(kd),length(object$smooth)) 
+      } else fit <- matrix(0,nrow(kd),length(object$smooth))
+      k <- 1; ## starting Xd
+      kk <- 1 ## starting col of fit for smooth terms
+    }
+    for (i in 1:length(object$smooth)) {
+      ii <- ts[k]:(ts[k]+dt[k]-1) ## index components for this term
+      ind <- object$smooth[[i]]$first.para:object$smooth[[i]]$last.para ## index coefs for this term
+      if (!is.null(object$dinfo$drop)) { 
+        drop <- object$dinfo$drop-object$smooth[[i]]$first.para+1
+        drop <- drop[drop<=length(ii)]
+      } else drop <- NULL
+      fit[,kk] <- Xbd(Xd[ii],object$coefficients[ind],kd[,ii,drop=FALSE],1,dt[k],object$dinfo$v[k],
+                        object$dinfo$qc[k],drop=drop)
+      if (se) se.fit[,kk] <- diagXVXd(Xd[ii],object$Vp[ind,ind],kd[,ii,drop=FALSE],1,dt[k],object$dinfo$v[k],
+                        object$dinfo$qc[k],drop=drop,n.threads=n.threads)^.5
+      k <-  k + 1; kk <- kk + 1
+    } 
+    fit.names <- c(if (se) colnames(pp$fit) else colnames(pp),unlist(lapply(object$smooth,function(x) x$label)))
+    colnames(fit) <- fit.names
+    if (se) { 
+      colnames(se.fit) <- fit.names
+      fit <- list(fit=fit,se.fit=se.fit)
+    }
+  } else if (type=="lpmatrix") {
+    fit <- Xbd(Xd,diag(length(object$coefficients)),kd,ts,dt,object$dinfo$v,object$dinfo$qc,drop=object$dinfo$drop)
+  } else { ## link or response
+    fit <- Xbd(Xd,object$coefficients,kd,ts,dt,object$dinfo$v,object$dinfo$qc,drop=object$dinfo$drop) + offset
+    if (type=="response") {
+      linkinv <- object$family$linkinv
+      dmu.deta <- object$family$mu.eta
+    } else linkinv <- dmu.deta <- NULL
+    if (se==TRUE) {
+      se.fit <- diagXVXd(Xd,object$Vp,kd,ts,dt,object$dinfo$v,object$dinfo$qc,drop=object$dinfo$drop,n.threads=n.threads)^.5
+      if (type=="response") {
+        se.fit <- se.fit * abs(dmu.deta(fit))
+        fit <- linkinv(fit)
+      }
+      fit <- list(fit=fit,se.fit=se.fit)
+    } else if (type=="response") fit <- linkinv(fit)
+  }
+  fit
+} ## end predict.bamd 
+
+
+
 
 sparse.model.matrix <- function(G,mf,chunk.size) {
 ## create a whole sparse model matrix
@@ -1472,12 +1729,32 @@ tero <- function(sm) {
   sm
 } ## tero
 
+AR.resid <- function(rsd,rho=0,AR.start=NULL) {
+## standardised residuals for AR1 model
+  if (rho==0) return(rsd)
+  ld <- 1/sqrt(1-rho^2) ## leading diagonal of root inverse correlation
+  sd <- -rho*ld         ## sub diagonal
+  N <- length(rsd)    
+  ## see rwMatrix() for how following are used...
+  ar.row <- c(1,rep(1:N,rep(2,N))[-c(1,2*N)]) ## index of rows to reweight
+  ar.weight <- c(1,rep(c(sd,ld),N-1))     ## row weights
+  ar.stop <- c(1,1:(N-1)*2+1)    ## (stop[i-1]+1):stop[i] are the rows to reweight to get ith row
+  if (!is.null(AR.start)) { ## need to correct the start of new AR sections...
+    ii <- which(AR.start==TRUE)
+    if (length(ii)>0) {
+          if (ii[1]==1) ii <- ii[-1] ## first observation does not need any correction
+          ar.weight[ii*2-2] <- 0 ## zero sub diagonal
+          ar.weight[ii*2-1] <- 1 ## set leading diagonal to 1
+    }
+  }
+  rwMatrix(ar.stop,ar.row,ar.weight,rsd)
+} ## AR.resid
 
 bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,na.action=na.omit,
                 offset=NULL,method="fREML",control=list(),scale=0,gamma=1,knots=NULL,sp=NULL,
                 min.sp=NULL,paraPen=NULL,chunk.size=10000,rho=0,AR.start=NULL,discrete=FALSE,
-                sparse=FALSE,cluster=NULL,
-                nthreads=NA,gc.level=1,use.chol=FALSE,samfrac=1,drop.unused.levels=TRUE,G=NULL,fit=TRUE,...)
+                sparse=FALSE,cluster=NULL,nthreads=NA,gc.level=1,use.chol=FALSE,samfrac=1,
+                drop.unused.levels=TRUE,G=NULL,fit=TRUE,...)
 
 ## Routine to fit an additive model to a large dataset. The model is stated in the formula, 
 ## which is then interpreted to figure out which bits relate to smooth terms and which to 
@@ -1488,6 +1765,7 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
 ## 'n.threads' is number of threads to use for non-cluster computation (e.g. combining 
 ## results from cluster nodes). If 'NA' then is set to max(1,length(cluster)).
 { control <- do.call("gam.control",control)
+  if (control$trace) t3 <- t2 <- t1 <- t0 <- proc.time()
   if (is.null(G)) { ## need to set up model!
     if (is.character(family))
             family <- eval(parse(text = family))
@@ -1525,15 +1803,17 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
     gp <- interpret.gam(formula) # interpret the formula 
     if (discretize) { 
       ## re-order the tensor terms for maximum efficiency, and 
-      ## signal that "re" terms should be constructed with marginals
+      ## signal that "re"/"fs" terms should be constructed with marginals
       ## also for efficiency
-      for (i in 1:length(gp$smooth.spec)) { 
+      if (length(gp$smooth.spec)>0) for (i in 1:length(gp$smooth.spec)) { 
         if (inherits(gp$smooth.spec[[i]],"tensor.smooth.spec")) 
         gp$smooth.spec[[i]] <- tero(gp$smooth.spec[[i]])
-        if (inherits(gp$smooth.spec[[i]],"re.smooth.spec")&&gp$smooth.spec[[i]]$dim>1) {
-          gp$smooth.spec[[i]]$xt <- "tensor"
-          class(gp$smooth.spec[[i]]) <- c("re.smooth.spec","tensor.smooth.spec")
+        if (inherits(gp$smooth.spec[[i]],c("re.smooth.spec","fs.smooth.spec"))&&gp$smooth.spec[[i]]$dim>1) {
+          #gp$smooth.spec[[i]]$xt <- "tensor"
+          class(gp$smooth.spec[[i]]) <- c(class(gp$smooth.spec[[i]]),"tensor.smooth.spec")
+                                        ##c("re.smooth.spec","tensor.smooth.spec")
           gp$smooth.spec[[i]]$margin <- list()
+          ## only ok for 'fs' with univariate metric variable (caught in 'fs' construcor)...
           for (j in 1:gp$smooth.spec[[i]]$dim) gp$smooth.spec[[i]]$margin[[j]] <- list(term=gp$smooth.spec[[i]]$term[j])
         }
       }
@@ -1554,7 +1834,12 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
    
     if (gc.level>0) gc()
 
-    mf <- eval(mf, parent.frame()) # the model frame now contains all the data 
+    mf <- eval(mf, parent.frame()) # the model frame now contains all the data
+   # if ("matrix"%in%unlist(lapply(mf,class))) {
+   #   mfattr <- attributes(mf)
+   #   mf <- lapply(mf,drop) # avoid single column matrices
+   #   mfattr -> attributes(mf)
+   # }
     if (nrow(mf)<2) stop("Not enough (non-NA) data to do anything meaningful")
     terms <- attr(mf,"terms")
     if (gc.level>0) gc()  
@@ -1583,19 +1868,31 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
       dk <- discrete.mf(gp,mf,pmf,m=discrete)
       mf0 <- dk$mf ## padded discretized model frame
       sparse.cons <- 0 ## default constraints required for tensor terms
+
     } else { 
       mf0 <- mini.mf(mf,chunk.size)
       if (sparse) sparse.cons <- 2 else sparse.cons <- -1
     }
     rm(pmf); ## no further use
-    
-    G <- gam.setup(gp,pterms=pterms,
+    if (control$trace) t1 <- proc.time()
+    reset <- TRUE
+    while (reset) {
+      G <- gam.setup(gp,pterms=pterms,
                  data=mf0,knots=knots,sp=sp,min.sp=min.sp,
                  H=NULL,absorb.cons=TRUE,sparse.cons=sparse.cons,select=FALSE,
                  idLinksBases=TRUE,scale.penalty=control$scalePenalty,
-                 paraPen=paraPen)
-  
+                 paraPen=paraPen,apply.by=!discretize)
+      if (!discretize&&ncol(G$X)>=chunk.size) { ## no point having chunk.size < p
+        chunk.size <- 4*ncol(G$X)
+        warning(gettextf("chunk.size < number of coefficients. Reset to %d",chunk.size))
+        if (chunk.size>=nrow(mf)) { ## no sense splitting up computation
+          mf0 <- mf ## just use full dataset
+        } else reset <- FALSE
+      } else reset <- FALSE
+    }
+    if (control$trace) t2 <- proc.time()
     if (discretize) {
+    
       v <- G$Xd <- list()
       ## have to extract full parametric model matrix from pterms and mf
       G$Xd[[1]] <- model.matrix(G$pterms,mf) 
@@ -1612,17 +1909,41 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
       drop <- rep(0,0) ## index of te related columns to drop
       for (i in 1:length(G$smooth)) {
         ts[kb] <- k
-        dt[kb] <- length(G$smooth[[i]]$margin)
-        if (inherits(G$smooth[[i]],"tensor.smooth")) {
-          if (inherits(G$smooth[[i]],"random.effect")&&!is.null(G$smooth[[i]]$rind)) {
+        ## first deal with any by variable (as first marginal of tensor)...
+        if (G$smooth[[i]]$by!="NA") {
+          dt[kb] <- 1
+          by.var <- dk$mf[[G$smooth[[i]]$by]][1:dk$nr[k]]
+          if (is.factor(by.var)) { 
+            ## create dummy by variable...
+            by.var <- as.numeric(by.var==G$smooth[[i]]$by.level)  
+          }
+          G$Xd[[k]] <- matrix(by.var,dk$nr[k],1)
+          #G$smooth[[i]]$by.kind <- k ## record which index vector relates to by variable
+          k <- k + 1
+        } else dt[kb] <- 0
+        ## ... by done
+        if (inherits(G$smooth[[i]],"tensor.smooth")) { 
+          nmar <- length(G$smooth[[i]]$margin) 
+          dt[kb] <- dt[kb] + nmar
+          if (inherits(G$smooth[[i]],"fs.interaction")&&which(G$smooth[[i]]$fterm==G$smooth[[i]]$term)!=1) {
+            ## have to reverse the terms because tensor representation assumes factor is first
+            G$smooth[[i]]$rind <- 2:1 ## (k+1):k
+          }          
+          # if (inherits(G$smooth[[i]],"random.effect")&&!is.null(G$smooth[[i]]$rind)) {
+          if (!is.null(G$smooth[[i]]$rind)) {
             ## terms re-ordered for efficiency, so the same has to be done on indices...
             rind <- k:(k+dt[kb]-1)    
             dk$nr[rind] <- dk$nr[k+G$smooth[[i]]$rind-1]
             G$kd[,rind] <- G$kd[,k+G$smooth[[i]]$rind-1]
-          }
-          
-          for (j in 1:dt[kb]) {
+          } 
+          #else if (inherits(G$smooth[[i]],"fs.interaction")&&which(G$smooth[[i]]$fterm==G$smooth[[i]]$term)!=1) {
+          #  ## have to reverse the terms because tensor representation assumes factor is first
+          #  dk$nr[k:(k+1)] <- dk$nr[(k+1):k]
+          #  G$kd[,k:(k+1)] <- G$kd[,(k+1):k]
+          #}          
+          for (j in 1:nmar) {
             G$Xd[[k]] <- G$smooth[[i]]$margin[[j]]$X[1:dk$nr[k],,drop=FALSE]
+            #G$smooth[[i]]$margin[[j]]$kind <- k ## record which index vector relates to this margin
             k <- k + 1 
           }
           ## deal with any side constraints on tensor terms  
@@ -1641,23 +1962,26 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
             v[[kb]] <- rep(0,0) ##
             if (!inherits(qrc,"character")||qrc!="no constraints") warning("unknown tensor constraint type")
           } 
-        } else {
+        } else { ## not a tensor smooth
           v[[kb]] <- rep(0,0)
-          dt[kb] <- 1
+          dt[kb] <- dt[kb] + 1
           G$Xd[[k]] <- G$X[1:dk$nr[k],G$smooth[[i]]$first.para:G$smooth[[i]]$last.para]
+          #G$smooth[[i]]$kind <- k ## which index vector relates to this smooth
           k <- k + 1
         }
         kb <- kb + 1
       }
-      if (length(drop>0)) G$drop <- drop
+      if (length(drop>0)) G$drop <- drop ## index of terms to drop as a result of side cons on tensor terms
+
       ## ... Xd is the list of discretized model matrices, or marginal model matrices
       ## kd contains indexing vectors, so the ith model matrix or margin is Xd[[i]][kd[i,],]
       ## ts[i] is the starting matrix in Xd for the ith model matrix, while dt[i] is the number 
-      ## of elements of Xd that make it up (1 for a dingleton, more for a tensor). 
+      ## of elements of Xd that make it up (1 for a singleton, more for a tensor). 
       ## v is list of Householder vectors encoding constraints and qc the constraint indicator.
       G$v <- v;G$ts <- ts;G$dt <- dt;G$qc <- qc
     } ## if (discretize)
 
+    if (control$trace) t3 <- proc.time()
     G$sparse <- sparse
 
     ## no advantage to "fREML" with no free smooths...
@@ -1678,7 +2002,7 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
 
     if (ncol(G$X)>nrow(mf)) stop("Model has more coefficients than data") 
   
-    if (ncol(G$X) > chunk.size) { ## no sense having chunk.size < p
+    if (ncol(G$X) > chunk.size && !discretize) { ## no sense having chunk.size < p
       chunk.size <- 4*ncol(G$X)
       warning(gettextf("chunk.size < number of coefficients. Reset to %d",chunk.size))    
     }
@@ -1763,6 +2087,8 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
 
   if (gc.level>0) gc()
 
+  if (control$trace) t4 <- proc.time()
+
   if (control$trace) cat("Fit complete. Finishing gam object.\n")
 
   if (scale < 0) { object$scale.estimated <- TRUE;object$scale <- object$scale.est} else {
@@ -1816,15 +2142,19 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
   ## note that predict.gam assumes that it must be ok not to split the 
   ## model frame, if no new data supplied, so need to supply explicitly
   class(object) <- c("bam","gam","glm","lm")
-  if (!G$discretize) object$linear.predictors <- 
+  if (!G$discretize) { object$linear.predictors <- 
           as.numeric(predict.bam(object,newdata=object$model,block.size=chunk.size,cluster=cluster))
-  
+  } else { ## store discretization specific information to help with discrete prediction
+    object$dinfo <- list(gp=gp, v = G$v, ts = G$ts, dt = G$dt, qc = G$qc, drop = G$drop)
+  } 
   rm(G);if (gc.level>0) gc()
 
   object$fitted.values <- family$linkinv(object$linear.predictors)
    
   object$residuals <- sqrt(family$dev.resids(object$y,object$fitted.values,object$prior.weights)) * 
                       sign(object$y-object$fitted.values)
+  if (rho!=0) object$std.rsd <- AR.resid(object$residuals,rho,object$model$"(AR.start)")
+
   object$deviance <- sum(object$residuals^2)
   object$aic <- family$aic(object$y,1,object$fitted.values,object$prior.weights,object$deviance) +
                 2*sum(object$edf)
@@ -1835,7 +2165,13 @@ bam <- function(formula,family=gaussian(),data=list(),weights=NULL,subset=NULL,n
   }
   environment(object$formula) <- environment(object$pred.formula) <-
   environment(object$terms) <- environment(object$pterms) <- 
-  environment(attr(object$model,"terms"))  <- .GlobalEnv
+  environment(attr(object$model,"terms"))  <- .GlobalEnv  
+  if (control$trace) { 
+    t5 <- proc.time()
+    t5 <- rbind(t1-t0,t2-t1,t3-t2,t4-t3,t5-t4)[,1:3]
+    row.names(t5) <- c("initial","gam.setup","pre-fit","fit","finalise")
+    print(t5)
+  }
   object
 } ## end of bam
 
