@@ -108,7 +108,7 @@ void tensorXty(double *Xy,double *work,double *work1, double *y,double *X,
     j = *dt - 1; 
     tensorXj(work,X,m,p,&j,k,n,&i);
     /* now form M'work */
-    singleXty(Xy+i*pd,work1,work,M,m + *dt-1,&pd, k + *n * (*dt-1),n);
+    singleXty(Xy+i*pd,work1,work,M,m + *dt-1,&pd, k + (ptrdiff_t) *n * (*dt-1),n);
   }
 } /* tensorXty */
 
@@ -135,23 +135,23 @@ void tensorXb(double *f,double *X, double *C,double *work, double *beta,
    no constraint is applied. 
 */ 
   char trans='N';
-  int pb=1,md,*kp,*kd,pd,i,j,one=1;
+  int pb=1,md,*kp,*kd,pd,i,j;
   double *M,done=1.0,dzero=0.0,*p0,*p1,*p2,*p3,*pf,*pc,x;
   M = X;
   for (i=0;i<*dt-1;i++) {
     pb *= p[i];
-    M += m[i]*p[i]; /* final marginal */ 
+    M += m[i] * (ptrdiff_t) p[i]; /* final marginal */ 
   }
   md = m[*dt - 1];
   pd = p[*dt -1];
-  kd = k + (*dt-1) * *n;
+  kd = k + (*dt-1) * (ptrdiff_t) *n;
   /* form work = M B, where vec(B) = beta */
   if (*qc<=0) { /* no constraint supplied */
     F77_CALL(dgemm)(&trans,&trans,&md,&pb, &pd, &done,
 		    M,&md,beta,&pd,&dzero,C,&md); 
   } else { /* their is a constraint matrix */
     /* first map supplied beta to unconstrained parameterization */ 
-    j = pb*pd; /* total number of coeffs - length of unconstrained beta */
+    j = pb * pd; /* total number of coeffs - length of unconstrained beta */
     
     *work = 0.0;x=0.0;
     for (p0=work+1,p1=p0+j-1,p2=beta,p3=v+1;p0<p1;p0++,p2++,p3++) { 
@@ -179,24 +179,26 @@ void tensorXb(double *f,double *X, double *C,double *work, double *beta,
 } /* tensorXb */
 
 void Xbd(double *f,double *beta,double *X,int *k, int *m,int *p, int *n, 
-	int *nx, int *ts, int *dt, int *nt,double *v,int *qc) {
-/* Forms f = X beta for X stored in the packed form described in function XWX */
-  int i,j,q,*pt,*off,*voff,*tps,dC=0,c1;
-  double *f0,*pf,*p0,*p1,*p2,*C,*work,maxp=0;
+	 int *nx, int *ts, int *dt, int *nt,double *v,int *qc,int *bc) {
+/* Forms f = X beta for X stored in the packed form described in function XWX
+   bc is number of cols of beta and f... 
+ */
+  ptrdiff_t *off,*voff;
+  int i,j,q,*pt,*tps,dC=0,c1;
+  double *f0,*pf,*p0,*p1,*p2,*C=NULL,*work,maxp=0;
   /* obtain various indices */
-  pt = (int *) R_chk_calloc((size_t)*nt,sizeof(int)); /* the term dimensions */
-  off = (int *) R_chk_calloc((size_t)*nx+1,sizeof(int)); /* offsets for X submatrix starts */
-  voff = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* offsets for v subvector starts */
-  tps = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* the term starts in param vector or XWy */
+  pt = (int *) CALLOC((size_t)*nt,sizeof(int)); /* the term dimensions */
+  off = (ptrdiff_t *) CALLOC((size_t)*nx+1,sizeof(ptrdiff_t)); /* offsets for X submatrix starts */
+  voff = (ptrdiff_t *) CALLOC((size_t)*nt+1,sizeof(ptrdiff_t)); /* offsets for v subvector starts */
+  tps = (int *) CALLOC((size_t)*nt+1,sizeof(int)); /* the term starts in param vector or XWy */
   for (q=i=0;i< *nt; i++) { /* work through the terms */
     for (j=0;j<dt[i];j++,q++) { /* work through components of each term */
-      off[q+1] = off[q] + p[q]*m[q]; /* submatrix start offsets */
+      off[q+1] = off[q] + p[q] * (ptrdiff_t) m[q]; /* submatrix start offsets */
       if (j>0 && j==dt[i]-1) {
-        c1 = pt[i]*m[q]; 
+        c1 = pt[i] * (ptrdiff_t) m[q]; 
         if (c1>dC) dC = c1; /* dimension of working matrix C */
       }
       if (j==0) pt[i] = p[q]; else pt[i] *= p[q]; /* term dimension */
-      //if (maxm<m[q]) maxm=m[q];
     } 
     if (qc[i]>0) voff[i+1] = voff[i] + pt[i]; else voff[i+1] = voff[i]; /* start of ith v matrix */
     if (maxp < pt[i]) maxp = pt[i];
@@ -204,39 +206,90 @@ void Xbd(double *f,double *beta,double *X,int *k, int *m,int *p, int *n,
     else tps[i+1] = tps[i] + pt[i] - 1; /* there is a tensor constraint to apply - reducing param count*/
   }
   /* now form the product term by term... */
-  pf=f0 = (double *)calloc((size_t)*n,sizeof(double));
+  pf=f0 = (double *)CALLOC((size_t)*n,sizeof(double));
   i = *n; if (i<maxp) i=maxp;
-  work = (double *)calloc((size_t)i,sizeof(double));
-  if (dC) C = (double *)calloc((size_t)dC,sizeof(double));
-  for (i=0;i < *nt;i++) { /* work through terms */ 
-    if (i==0) f0 = f; /* result written straight to f for i==0 */
-    if (dt[i]==1) singleXb(f0,work,X+off[ts[i]],beta+tps[i],k + *n *ts[i],m+ts[i], p+ts[i],n); 
-    else tensorXb(f0,X+off[ts[i]],C,work, beta+tps[i],
-                  m+ts[i], p+ts[i],dt+i,k + *n *ts[i],n,v+voff[i], qc+i);
-    if (i>0) {
-      for (p0=f,p1=f + *n,p2=f0;p0<p1;p0++,p2++) *p0 += *p2; /* f <- f + f0 */     
-    } else { f0=pf; /* restore f0 */}
+  work = (double *)CALLOC((size_t)i,sizeof(double));
+  if (dC) C = (double *)CALLOC((size_t)dC,sizeof(double));
+  for (j=0;j < *bc;j++) {
+    for (i=0;i < *nt;i++) { /* work through terms */ 
+      if (i==0) f0 = f; /* result written straight to f for i==0 */
+      if (dt[i]==1) singleXb(f0,work,X+off[ts[i]],beta+tps[i],k + (ptrdiff_t) *n * ts[i],m+ts[i], p+ts[i],n); 
+      else tensorXb(f0,X+off[ts[i]],C,work, beta+tps[i],
+		    m+ts[i], p+ts[i],dt+i,k + (ptrdiff_t) *n * ts[i],n,v+voff[i], qc+i);
+      if (i>0) {
+        for (p0=f,p1=f + *n,p2=f0;p0<p1;p0++,p2++) *p0 += *p2; /* f <- f + f0 */     
+      } else { f0=pf; /* restore f0 */}
+    }
+    f += *n;beta += tps[*nt]; /* move on to next column */
   }
-  if (dC) R_chk_free(C);
-  R_chk_free(work);R_chk_free(f0);
-  R_chk_free(pt);R_chk_free(off);R_chk_free(voff);R_chk_free(tps);
+  if (dC) FREE(C);
+  FREE(work);FREE(f0);
+  FREE(pt);FREE(off);FREE(voff);FREE(tps);
 } /* Xb */
 
+void diagXVXt(double *diag,double *V,double *X,int *k,int *m,int *p, int *n, 
+	      int *nx, int *ts, int *dt, int *nt,double *v,int *qc,int *pv,int *nthreads) {
+/* Forms diag(XVX') where X is stored in the compact form described in XWXd.
+   V is a pv by pv matrix. 
+   Parallelization is by splitting the columns of V into nthreads subsets.
+   Currently inefficient. Could be speeded up by a factor of 2, by supplying a
+   square root of V in place of V.
+*/
+  double *xv,*dc,*p0,*p1,*p2,*p3,*ei,*xi;
+  ptrdiff_t bsj,bs,bsf,i,j,kk;
+  int one=1;
+  #ifndef SUPPORT_OPENMP
+  *nthreads = 1;
+  #endif
+  if (*nthreads<1) *nthreads = 1;
+  if (*nthreads > *pv) *nthreads = *pv;
+  xv = (double *) CALLOC((size_t) *nthreads * *n,sizeof(double)); /* storage for cols of XV */
+  xi = (double *) CALLOC((size_t) *nthreads * *n,sizeof(double)); /* storage for cols of X */
+  ei = (double *) CALLOC((size_t) *nthreads * *pv,sizeof(double)); /* storage for identity matrix cols */
+  dc = (double *) CALLOC((size_t) *nthreads * *n,sizeof(double)); /* storage for components of diag */
+  if (*nthreads>1) {
+    bs = *pv / *nthreads;
+    while (bs * *nthreads < *pv) bs++;
+    while (bs * *nthreads - bs >= *pv) (*nthreads)--;
+    bsf = *pv - (bs * *nthreads - bs); 
+  } else {
+    bsf = bs = *pv;
+  }
+  #ifdef SUPPORT_OPENMP
+  #pragma omp parallel for private(j,bsj,i,kk,p0,p1,p2,p3) num_threads(*nthreads)
+  #endif  
+  for (j=0;j < *nthreads;j++) {
+    if (j == *nthreads - 1) bsj = bsf; else bsj = bs;
+    for (i=0;i<bsj;i++) { /* work through this block's columns */
+      kk = j * bs + i;
+      ei[j * *pv + kk] = 1;if (i>0) ei[j * *pv + kk - 1] = 0;
+      Xbd(xv + j * *n,V + kk * *pv,X,k,m,p,n,nx,ts,dt,nt,v,qc,&one); /* XV[:,kk] */
+      Xbd(xi + j * *n,ei + j * *pv,X,k,m,p,n,nx,ts,dt,nt,v,qc,&one); /* X[:,kk] inefficient, but deals with constraint*/
+      p0 = xi + j * *n;p1=xv + j * *n;p2 = dc + j * *n;p3 = p2 + *n;
+      for (;p2<p3;p0++,p1++,p2++) *p2 += *p0 * *p1; /* elementwise product of XV[:,kk] X[:,kk] */
+    } 
+  } /* parallel loop end */
+  /* sum the contributions from the different threads into diag... */
+  for (p0=diag,p1=p0+ *n,p2=dc;p0<p1;p0++,p2++) *p0 = *p2;
+  for (i=1;i< *nthreads;i++) for (p0=diag,p1=p0+ *n;p0<p1;p0++,p2++) *p0 += *p2;
+  FREE(xv);FREE(dc);FREE(xi);FREE(ei);
+} /* diagXVXt */
+
 
 void XWyd(double *XWy,double *y,double *X,double *w,int *k, int *m,int *p, int *n, 
          int *nx, int *ts, int *dt, int *nt,double *v,int *qc,
          int *ar_stop,int *ar_row,double *ar_weights) {
-  double *Wy,*p0,*p1,*p2,*p3,done=1.0,dzero=0.0,*Xy0,*work,*work1,x;
-  int q,i,j,*pt,*off,*voff,*tps,maxm=0,maxp=0,one=1,zero=0;
-  char trans='T';
+  double *Wy,*p0,*p1,*p2,*p3,*Xy0,*work,*work1,x;
+  ptrdiff_t q,i,j,*off,*voff;
+  int *tps,maxm=0,maxp=0,one=1,zero=0,*pt;
   if (*ar_stop>=0) { /* model has AR component, requiring sqrt(weights) */
     for (p0 = w,p1 = w + *n;p0<p1;p0++) *p0 = sqrt(*p0);
   }
   /* obtain various indices */
-  pt = (int *) R_chk_calloc((size_t)*nt,sizeof(int)); /* the term dimensions */
-  off = (int *) R_chk_calloc((size_t)*nx+1,sizeof(int)); /* offsets for X submatrix starts */
-  voff = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* offsets for v subvector starts */
-  tps = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* the term starts in param vector or XWy */
+  pt = (int *) CALLOC((size_t)*nt,sizeof(int)); /* the term dimensions */
+  off = (ptrdiff_t *) CALLOC((size_t)*nx+1,sizeof(ptrdiff_t)); /* offsets for X submatrix starts */
+  voff = (ptrdiff_t *) CALLOC((size_t)*nt+1,sizeof(ptrdiff_t)); /* offsets for v subvector starts */
+  tps = (int *) CALLOC((size_t)*nt+1,sizeof(int)); /* the term starts in param vector or XWy */
   for (q=i=0;i< *nt; i++) { /* work through the terms */
     for (j=0;j<dt[i];j++,q++) { /* work through components of each term */
       off[q+1] = off[q] + p[q]*m[q]; /* submatrix start offsets */
@@ -248,37 +301,36 @@ void XWyd(double *XWy,double *y,double *X,double *w,int *k, int *m,int *p, int *
     if (qc[i]<=0) tps[i+1] = tps[i] + pt[i]; /* where ith terms starts in param vector */ 
     else tps[i+1] = tps[i] + pt[i] - 1; /* there is a tensor constraint to apply - reducing param count*/
   }
-  Xy0 =  (double *) R_chk_calloc((size_t)maxp,sizeof(double));
-  work =  (double *) R_chk_calloc((size_t)*n,sizeof(double));
-  work1 = (double *) R_chk_calloc((size_t)maxm,sizeof(double));
+  Xy0 =  (double *) CALLOC((size_t)maxp,sizeof(double));
+  work =  (double *) CALLOC((size_t)*n,sizeof(double));
+  work1 = (double *) CALLOC((size_t)maxm,sizeof(double));
   /* apply W to y - for AR models this needs to be expanded */
-  Wy = (double *) R_chk_calloc((size_t)*n,sizeof(double)); /* Wy */
+  Wy = (double *) CALLOC((size_t)*n,sizeof(double)); /* Wy */
   for (p0=Wy,p1=Wy + *n,p2=w;p0<p1;p0++,y++,p2++) *p0 = *y * *p2; 
   if (*ar_stop>=0) { /* AR components present (weights are sqrt, therefore) */
-    rwMatrix(ar_stop,ar_row,ar_weights,Wy,n,&one,&zero);
-    rwMatrix(ar_stop,ar_row,ar_weights,Wy,n,&one,&one); /* transpose of transform applied */
+    rwMatrix(ar_stop,ar_row,ar_weights,Wy,n,&one,&zero,work);
+    rwMatrix(ar_stop,ar_row,ar_weights,Wy,n,&one,&one,work); /* transpose of transform applied */
     for (p0=w,p1=w + *n,p2=Wy;p0<p1;p0++,p2++) *p2 *= *p0; /* sqrt weights again */
   }
   /* now loop through terms applying the components of X'...*/
   for (i=0;i<*nt;i++) {
     if (dt[i]>1) { /* it's a tensor */
-      tensorXty(Xy0,work,work1,Wy,X+off[ts[i]],m+ts[i],p+ts[i],dt+i,k+ts[i] * *n,n);
+      tensorXty(Xy0,work,work1,Wy,X+off[ts[i]],m+ts[i],p+ts[i],dt+i,k+ts[i] * (ptrdiff_t) *n,n);
       if (qc[i]>0) { /* there is a constraint to apply Z'Xy0: form Q'Xy0 and discard first row... */
         /* Q' = I - vv' */
         for (x=0.0,p0=Xy0,p1=p0 + pt[i],p2=v+voff[i];p0<p1;p0++,p2++) x += *p0 * *p2; /* x = v'Xy0 */
         p0=XWy + tps[i];p1 = p0 + pt[i]-1;p2 = v+voff[i] + 1;p3=Xy0+1;
         for (;p0<p1;p0++,p2++,p3++) *p0 = *p3 - x * *p2; /* (I-vv')Xy0 less first element */
-        //F77_CALL(dgemv)(&trans, pt+i, qc+i,&done,Q+Qoff[i],pt+i,Xy0,&one,&dzero,XWy + tps[i],&one);
       } else { /* straight copy */
         for (p0=Xy0,p1=p0+pt[i],p2=XWy+tps[i];p0<p1;p0++,p2++) *p2 = *p0;
       }
     } else { /* it's a singleton */
       //singleXty(double *Xy,double *temp,double *y,double *X, int *m,int *p, int *k, int *n)
-      singleXty(XWy+tps[i],work1,Wy,X+off[ts[i]], m+ts[i],p+ts[i], k+ts[i] * *n,n);
+      singleXty(XWy+tps[i],work1,Wy,X+off[ts[i]], m+ts[i],p+ts[i], k+ts[i] * (ptrdiff_t) *n,n);
     }
   }
-  R_chk_free(Wy); R_chk_free(Xy0); R_chk_free(work); R_chk_free(work1); 
-  R_chk_free(pt); R_chk_free(off); R_chk_free(voff); R_chk_free(tps);
+  FREE(Wy); FREE(Xy0); FREE(work); FREE(work1); 
+  FREE(pt); FREE(off); FREE(voff); FREE(tps);
 } /* XWy */
 
 void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx, int *ts, 
@@ -294,7 +346,7 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
    then the ith row of the corresponding full sub-matrix is Xj[kj[i],]. The blocks of Xt 
    need not be directly given by terms like Xj[kj[i],], some are actually given by row 
    tensors of several such matrices. Conceptually Xt has nt <= nx blocks (terms) and the ith 
-   block starts at sub-matrix  ts[i] of X, and is made up of the row tensor produce of 
+   block starts at sub-matrix  ts[i] of X, and is made up of the row tensor product of 
    dt[i] consecutive sub-matrices. 
     
    Tensor product terms may have constraint matrices Z, which post multiply the tensor product 
@@ -304,29 +356,29 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
    AR models are handled via the 3 ar_* arrays. ar_stop[0] < 0 signals no AR. 
   
 */  
-  int r,c,i,j,q,*pt,*pd,*off,a,b,*tps,ptot,maxp=0,maxm=0,*voff,pa,pb,kk,dk,rk,*start,one=1,zero=0; 
-  double done=1.0,dzero=0.0,*p0,*p1,*p2, *Xi,*temp,*tempn,*xwx,*xwx0,
+  int r,c,i,j,q,*pt,*pd,a,b,*tps,ptot,maxp=0,maxm=0,pa,pb,kk,dk,*start,one=1,zero=0; 
+  ptrdiff_t *off,*voff;
+  double *p0,*p1,*p2, *Xi,*temp,*tempn,*xwx,*xwx0,
     *XiB,*tempB,*tempnB,*x0,*x1,x;
-  char trans='T',not_trans='N';
   #ifndef SUPPORT_OPENMP
   *nthreads = 1;
   #endif
   if (*nthreads<1) *nthreads = 1;
-  start = (int *) R_chk_calloc((size_t)(*nthreads+1),sizeof(int));
-  XiB = (double *) R_chk_calloc((size_t)(*n * *nthreads),sizeof(double)); /* column of X */
-  tempnB = (double *) R_chk_calloc((size_t)(*n * *nthreads),sizeof(double));
-  pt = (int *) R_chk_calloc((size_t)*nt,sizeof(int)); /* the term dimensions */
-  pd = (int *) R_chk_calloc((size_t)*nt,sizeof(int)); /* storage for last marginal size */
-  off = (int *) R_chk_calloc((size_t)*nx+1,sizeof(int)); /* offsets for X submatrix starts */
-  voff = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* offsets for v subvectors starts */
-  tps = (int *) R_chk_calloc((size_t)*nt+1,sizeof(int)); /* the term starts */
+  start = (int *) CALLOC((size_t)(*nthreads+1),sizeof(int));
+  XiB = (double *) CALLOC((size_t)(*n * *nthreads),sizeof(double)); /* column of X */
+  tempnB = (double *) CALLOC((size_t)(*n * *nthreads),sizeof(double));
+  pt = (int *) CALLOC((size_t)*nt,sizeof(int)); /* the term dimensions */
+  pd = (int *) CALLOC((size_t)*nt,sizeof(int)); /* storage for last marginal size */
+  off = (ptrdiff_t *) CALLOC((size_t)*nx+1,sizeof(ptrdiff_t)); /* offsets for X submatrix starts */
+  voff = (ptrdiff_t *) CALLOC((size_t)*nt+1,sizeof(ptrdiff_t)); /* offsets for v subvectors starts */
+  tps = (int *) CALLOC((size_t)*nt+1,sizeof(int)); /* the term starts */
   if (*ar_stop>=0) { /* model has AR component, requiring sqrt(weights) */
     for (p0 = w,p1 = w + *n;p0<p1;p0++) *p0 = sqrt(*p0);
   }
   for (q=i=0;i< *nt; i++) { /* work through the terms */
     for (j=0;j<dt[i];j++,q++) {
       if (j==dt[i]-1) pd[i] = p[q]; /* the relevant dimension for deciding product ordering */
-      off[q+1] = off[q] + p[q]*m[q]; /* submatrix start offsets */
+      off[q+1] = off[q] + p[q] * (ptrdiff_t) m[q]; /* submatrix start offsets */
       if (j==0) pt[i] = p[q]; else pt[i] *= p[q]; /* term dimension */
       if (maxm<m[q]) maxm=m[q];
     } 
@@ -335,9 +387,9 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
     if (qc[i]<=0) tps[i+1] = tps[i] + pt[i]; /* where ith terms starts in param vector */ 
     else tps[i+1] = tps[i] + pt[i] - 1; /* there is a tensor constraint to apply - reducing param count*/
   }
-  tempB = (double *) R_chk_calloc((size_t)(maxm * *nthreads),sizeof(double));
-  xwx = (double *) R_chk_calloc((size_t)(maxp*maxp),sizeof(double)); /* working cross product storage */
-  xwx0 = (double *) R_chk_calloc((size_t)(maxp*maxp),sizeof(double)); /* working cross product storage */
+  tempB = (double *) CALLOC((size_t) maxm * *nthreads,sizeof(double));
+  xwx = (double *) CALLOC((size_t) maxp * maxp,sizeof(double)); /* working cross product storage */
+  xwx0 = (double *) CALLOC((size_t) maxp * maxp,sizeof(double)); /* working cross product storage */
   ptot = tps[*nt]; /* total number of parameters */
   for (r=0;r < *nt;r++) for (c=r;c< *nt;c++) { /* the block loop */
     if (pd[r]>pd[c]) { /* Form Xr'WXc */
@@ -346,7 +398,7 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
       a=c;b=r; 
     }
     /* split cols between threads... */  
-    dk = pt[b] / *nthreads; rk = pt[b] % *nthreads;
+    dk = pt[b] / *nthreads; //rk = pt[b] % *nthreads;
     if (dk * *nthreads < pt[b]) dk++;start[0]=0; 
     for (i=0;i<*nthreads;i++) { 
       start[i+1] = start[i] + dk;
@@ -361,31 +413,31 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
       #endif
       for (kk=0;kk<*nthreads;kk++) { 
         /* allocate thread specific storage... */
-        temp = tempB + kk * maxm;
-        Xi = XiB + kk * *n;
-        tempn = tempnB + kk * *n;
+        temp = tempB + kk * (ptrdiff_t) maxm;
+        Xi = XiB + kk * (ptrdiff_t) *n;
+        tempn = tempnB + kk * (ptrdiff_t) *n;
         for (i=start[kk];i<start[kk+1];i++) { /* loop over columns of Xb */ 
           /* extract Xb[:,i]... */
           if (dt[b]>1) { /* tensor */
             for (p0=Xi,p1=p0+*n;p0<p1;p0++) *p0 = 1.0; /* set Xi to 1 */
             tensorXj(Xi,X+off[ts[b]], m + ts[b] , p + ts[b], dt+b, 
-                     k + *n * ts[b], n,&i);
+                     k + (ptrdiff_t) *n * ts[b], n,&i);
           } else { /* singleton */
-            singleXj(Xi,X+off[ts[b]], m + ts[b], k + *n * ts[b], n,&i);
+            singleXj(Xi,X+off[ts[b]], m + ts[b], k + (ptrdiff_t)*n * ts[b], n,&i);
           }
           /* apply W to Xi - for AR models this is sqrt(W) */
           for (p0=w,p1=w + *n,p2=Xi;p0<p1;p0++,p2++) *p2 *= *p0; 
           if (*ar_stop>=0) { /* AR components present (weights are sqrt, therefore) */
-            rwMatrix(ar_stop,ar_row,ar_weights,Xi,n,&one,&zero);
-            rwMatrix(ar_stop,ar_row,ar_weights,Xi,n,&one,&one); /* transpose of transform applied */
+            rwMatrix(ar_stop,ar_row,ar_weights,Xi,n,&one,&zero,tempn);
+            rwMatrix(ar_stop,ar_row,ar_weights,Xi,n,&one,&one,tempn); /* transpose of transform applied */
             for (p0=w,p1=w + *n,p2=Xi;p0<p1;p0++,p2++) *p2 *= *p0; /* sqrt weights again */
           }
           /* now form Xa'WXb[:,i]... */  
           if (dt[a]>1) { /* tensor */
             tensorXty(xwx + i * pt[a],tempn,temp,Xi,X+off[ts[a]],m+ts[a],p+ts[a],
-                      dt+a,k+ *n * ts[a], n);
+                      dt+a,k+ (ptrdiff_t)*n * ts[a], n);
           } else { /* singleton */
-            singleXty(xwx + i * pt[a],temp,Xi,X+off[ts[a]],m+ts[a],p+ts[a],k + *n * ts[a],n);
+            singleXty(xwx + i * pt[a],temp,Xi,X+off[ts[a]],m+ts[a],p+ts[a],k + (ptrdiff_t)*n * ts[a],n);
           }  
         } /* loop over columns of Xb */
       }
@@ -394,12 +446,6 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
 
     /* if Xb is tensor, may need to apply constraint */
     if (dt[a]>1&&qc[a]>0) { /* first term is a tensor with a constraint */
-      ///* copy xwx to xwx0 */
-      //for (p0=xwx,p1=p0 + pt[b]*pt[a],p2=xwx0;p0<p1;p0++,p2++) *p2 = *p0;
-      /* apply Qa' */
-      //F77_CALL(dgemm)(&trans,&not_trans,qc+a,pt+b,pt+a, &done,
-      //		    Q+Qoff[a],pt+a,xwx0,pt+a,&dzero,xwx,qc+a);
-      //pa = qc[a]; /* rows of xwx */
       x0=x1=xwx; /* pointers to columns of xwx */ 
       /* col by col form (I-vv')xwx, dropping first row... */
       for (j=0;j<pt[b];j++) {
@@ -410,7 +456,7 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
     } else pa = pt[a];
     if (dt[b]>1&&qc[b]>0) { /* second term is a tensor with a constraint */
       /* copy xwx to xwx0 */
-      for (p0=xwx,p1=p0 + pt[b]*pa,p2=xwx0;p0<p1;p0++,p2++) *p2 = *p0;
+      for (p0=xwx,p1=p0 + pt[b] * (ptrdiff_t) pa,p2=xwx0;p0<p1;p0++,p2++) *p2 = *p0;
       /* row by row form xwx(I-vv') dropping first col... */
       for (j=0;j<pa;j++) {
         for (x=0.0,p0=xwx0+j,p1=v+voff[b],p2=p1+pt[b];p1<p2;p0 += pa,p1++) x += *p0 * *p1; 
@@ -418,24 +464,21 @@ void XWXd(double *XWX,double *X,double *w,int *k, int *m,int *p, int *n, int *nx
         for (p0=xwx0+j+pa,p1=v+voff[b],p2=p1+pt[b],p1++;p1<p2;x1+=pa,p1++,p0+=pa) *x1 = *p0 - *p1 *x; 
       }
       pb=pt[b]-1;
-      ///* apply Qa from right */
-      //F77_CALL(dgemm)(&not_trans,&not_trans,&pa,qc+b,pt+b, &done,
-      //		      xwx0,&pa, Q+Qoff[b],pt+b,&dzero,xwx,&pa);
-      //pb = qc[b]; /* rows of xwx */
     } else pb=pt[b];
     /* copy result into overall XWX*/  
   
     if (pd[r]>pd[c]) { /* xwx = Xr'WXc */
       for (i=0;i<pa;i++) for (j=0;j<pb;j++) 
-      XWX[i+tps[r]+(j+tps[c])*ptot] = XWX[j+tps[c]+(i+tps[r])*ptot] = xwx[i + pa * j];
+      XWX[i+tps[r]+(j+tps[c]) * (ptrdiff_t)ptot] = XWX[j+tps[c]+(i+tps[r]) * (ptrdiff_t)ptot] 
+                                                 = xwx[i + pa * (ptrdiff_t)j];
     } else { /* Form xwx = Xc'WXr */
       for (i=0;i<pa;i++) for (j=0;j<pb;j++) 
-      XWX[i+tps[c]+(j+tps[r])*ptot] = XWX[j+tps[r]+(i+tps[c])*ptot] = xwx[i + pa * j];
+      XWX[i+tps[c]+(j+tps[r])*(ptrdiff_t)ptot] = XWX[j+tps[r]+(i+tps[c])*(ptrdiff_t)ptot] 
+                                               = xwx[i + pa * (ptrdiff_t)j];
     }
   } /* end of block loop */
-  R_chk_free(start);
-  R_chk_free(XiB); R_chk_free(tempnB); R_chk_free(pt); R_chk_free(pd); R_chk_free(off);
-  R_chk_free(voff); R_chk_free(tps); R_chk_free(tempB); R_chk_free(xwx); R_chk_free(xwx0);
-  //R_chk_free(); R_chk_free(); R_chk_free(); R_chk_free(); R_chk_free();
+  FREE(start);
+  FREE(XiB); FREE(tempnB); FREE(pt); FREE(pd); FREE(off);
+  FREE(voff); FREE(tps); FREE(tempB); FREE(xwx); FREE(xwx0);
 } /* XWX */
 
