@@ -17,6 +17,15 @@
 
 //#define OMP_REPORT // define to have all routines using omp report on start and end.
 
+/* sed -i 's/old-text/new-text/g' *.c
+   is quite useful!!
+*/
+// For safe memory handling from R...
+#define CALLOC R_chk_calloc
+#define FREE R_chk_free
+// Can reset to check for memory errors...
+//#define CALLOC calloc
+//#define FREE free
 void magic(double *y,double *X,double *sp0,double *def_sp,double *S,double *H,double *L,
 	   double *lsp0,double *gamma,double *scale, int *control,int *cS,double *rank_tol,
 	   double *tol,double *b,double *rV,double *norm_const,int *n_score,int *nt);
@@ -31,7 +40,7 @@ void gdi1(double *X,double *E,double *Es,double *rS,double *U1,
 	  int *REML,int *fisher,int *fixed_penalty,int *nthreads);     
 
 void gdi2(double *X,double *E,double *Es,double *rS,double *U1,
-	  double *sp,double *theta,double *z,double *w,double *wf,
+	  double *sp,double *theta,double *z,double *w,double *wz,double *wf,
           double *Dth,double *Det,double *Det2,double *Dth2,double *Det_th,
           double *Det2_th,double *Det3,double *Det_th2,
           double *Det4, double *Det3_th, double *Det2_th2,
@@ -39,10 +48,10 @@ void gdi2(double *X,double *E,double *Es,double *rS,double *U1,
           double *ldet, double *ldet1,double *ldet2,double *rV,
           double *rank_tol,int *rank_est,
 	  int *n,int *q, int *M,int *n_theta, int *Mp,int *Enrow,int *rSncol,int *deriv,
-	  int *fixed_penalty,int *nt);
+	  int *fixed_penalty,int *nt,int *type);
 
-void pls_fit1(double *y,double *X,double *w,double *E,double *Es,int *n,int *q,int *rE,double *eta,
-	      double *penalty,double *rank_tol,int *nt);
+void pls_fit1(double *y,double *X,double *w,double *wy,double *E,double *Es,int *n,int *q,int *rE,double *eta,
+	      double *penalty,double *rank_tol,int *nt,int *use_wy);
 
 void get_detS2(double *sp,double *sqrtS, int *rSncol, int *q,int *M, int * deriv, 
                double *det, double *det1, double *det2, double *d_tol,
@@ -76,7 +85,9 @@ void XWyd(double *XWy,double *y,double *X,double *w,int *k, int *m,int *p, int *
 	  int *nx, int *ts, int *dt, int *nt,double *v,int *qc,
           int *ar_stop,int *ar_row,double *ar_weights);
 void Xbd(double *f,double *beta,double *X,int *k, int *m,int *p, int *n, 
-	 int *nx, int *ts, int *dt, int *nt,double *v,int *qc);
+	 int *nx, int *ts, int *dt, int *nt,double *v,int *qc,int *bc);
+void diagXVXt(double *diag,double *V,double *X,int *k,int *m,int *p, int *n, 
+	      int *nx, int *ts, int *dt, int *nt,double *v,int *qc,int *pv,int *nthreads);
 
 /* various service routines */
 
@@ -84,7 +95,7 @@ void tweedious(double *w,double *w1,double *w2, double *w1p,double *w2p,double *
 	       double *y,double *eps,int *n,
                double *th,double *rho,double *a, double *b);
 void psum(double *y, double *x,int *index,int *n);
-void rwMatrix(int *stop,int *row,double *w,double *X,int *n,int *p,int *trans);
+void rwMatrix(int *stop,int *row,double *w,double *X,int *n,int *p,int *trans,double *work);
 void in_out(double *bx, double *by, double *break_code, double *x,double *y,int *in, int *nb, int *n);
 void Rlanczos(double *A,double *U,double *D,int *n, int *m, int *lm,double *tol,int *nt);
 void RuniqueCombs(double *X,int *ind,int *r, int *c);
@@ -101,8 +112,8 @@ void mgcv_chol(double *a,int *pivot,int *n,int *rank);
 void mgcv_svd(double *x,double *u, double *d,int *r,int *c);
 void mgcv_qrqy(double *b,double *a,double *tau,int *r,int *c,int *k,int *left,int *tp);
 void mgcv_qrqy0(double *b,double *a,double *tau,int *r,int *c,int *k,int *left,int *tp);
-void mgcv_backsolve(double *R,int *r,int *c,double *B,double *C, int *bc);
-void mgcv_forwardsolve(double *R,int *r,int *c,double *B,double *C, int *bc);
+void mgcv_backsolve(double *R,int *r,int *c,double *B,double *C, int *bc, int *right);
+void mgcv_forwardsolve(double *R,int *r,int *c,double *B,double *C, int *bc, int *right);
 void mgcv_qr(double *x, int *r, int *c,int *pivot,double *tau);
 void mgcv_qr2(double *x, int *r, int *c,int *pivot,double *tau);
 void update_qr(double *Q,double *R,int *n, int *q,double *lam, int *k);
